# Licensed to CRATE Technology GmbH ("Crate") under one or more contributor
# license agreements.  See the NOTICE file distributed with this work for
# additional information regarding copyright ownership.  Crate licenses
# this file to you under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.  You may
# obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# License for the specific language governing permissions and limitations
# under the License.
#
# However, if you have executed another commercial license agreement
# with Crate these terms will supersede the license and you may use the
# software solely pursuant to the terms of the relevant commercial agreement.

import enum
from argparse import Namespace
from typing import Any, Callable, Dict, Optional, Tuple

import requests
from yarl import URL

from croud.config import CONFIG
from croud.printer import print_error, print_info

ResponsePair = Tuple[Optional[Dict], Optional[Dict]]


class RequestMethod(enum.Enum):
    DELETE = "delete"
    GET = "get"
    PATCH = "patch"
    POST = "post"
    PUT = "put"


def noop(*args, **kwargs):
    pass


class Client:
    """
    A client for CrateDB Cloud API requests
    """

    def __init__(
        self,
        endpoint: str,
        *,
        token: str = None,
        on_token: Callable = None,
        region: str = None,
        sudo: bool = False,
        _verify_ssl: bool = True,
    ):
        """
        :param str endpoint:
          CrateDB Cloud API endpoint
        :param str token:
          The authentication token used as session cookie
        :param Callable on_token:
          A setter method that is called when a new auth token is retrieved
        :param str region:
          Region to get data from or to which to deploy (defines the
          ``X-Region`` HTTP header value)
        :param bool sudo:
          Whether or not to make requests as superuser (defines the
          ``X-Auth-Sudo`` HTTP header value)
        :param bool _verify_ssl:
          A private variable that must only be used during tests!
        """

        self.base_url = URL(endpoint)
        self._token = token
        self._on_token = on_token or noop

        self.session = requests.Session()
        self.session.cookies["session"] = self._token
        if _verify_ssl is False:
            self.session.verify = False
        if region:
            self.session.headers["X-Region"] = region
        if sudo:
            self.session.headers["X-Auth-Sudo"] = "1"

    @staticmethod
    def from_args(args: Namespace) -> "Client":
        return Client(
            CONFIG.endpoint,
            token=CONFIG.token,
            on_token=CONFIG.set_current_auth_token,
            region=args.region,
            sudo=args.sudo,
        )

    def request(
        self,
        method: RequestMethod,
        endpoint: str,
        *,
        params: dict = None,
        body: dict = None,
    ):
        # When logging out, the Gateway may respond with a redirect in case the
        # session is still valid and the IDP identifier is present.
        # We need to follow that redirect in order to fully log out the user!
        kwargs: Dict[str, Any] = {
            "allow_redirects": "logout" in endpoint,
        }
        if params is not None:
            kwargs["params"] = params
        if body is not None:
            kwargs["json"] = body

        try:
            response = self.session.request(
                method.value, str(self.base_url.with_path(endpoint)), **kwargs
            )
        except requests.RequestException as e:
            message = (
                f"Failed to perform request on '{e.request.url}'. "
                f"Original error was: '{e}'"
            )
            return None, {"message": message, "success": False}

        if response.is_redirect:
            redirect_url = URL(response.headers["Location"])
            print_error("Unauthorized.")
            if redirect_url.query.get("rd", "") == endpoint:  # login redirect
                # When use is unauthorized, the API request returns a 302
                # redirect to `/?rd=<api_endpoint>`
                print_info("Use `croud login` to login to CrateDB Cloud.")
            else:
                print_error("Oops. Something unexpected happened.")
            exit(1)

        # Refresh a previously provided token because it has timed out
        response_token = response.cookies.get("session")
        if response_token and response_token != self._token:
            self._token = response_token
            self._on_token(response_token)

        return self.decode_response(response)

    def delete(
        self, endpoint: str, *, params: dict = None, body: dict = None
    ) -> ResponsePair:
        return self.request(RequestMethod.DELETE, endpoint, params=params, body=body)

    def get(self, endpoint: str, *, params: dict = None) -> ResponsePair:
        return self.request(RequestMethod.GET, endpoint, params=params)

    def patch(
        self, endpoint: str, *, params: dict = None, body: dict = None
    ) -> ResponsePair:
        return self.request(RequestMethod.PATCH, endpoint, params=params, body=body)

    def post(
        self, endpoint: str, *, params: dict = None, body: dict = None
    ) -> ResponsePair:
        return self.request(RequestMethod.POST, endpoint, params=params, body=body)

    def put(
        self, endpoint: str, *, params: dict = None, body: dict = None
    ) -> ResponsePair:
        return self.request(RequestMethod.PUT, endpoint, params=params, body=body)

    def decode_response(self, resp: requests.Response) -> ResponsePair:
        if resp.status_code == 204:
            # response is empty
            return None, None

        try:
            # API always returns JSON, unless there's an unhandled server error
            body = resp.json()
        except ValueError:
            body = {"message": f"{resp.status_code} - {resp.reason}", "success": False}

        if resp.status_code >= 400:
            return None, body
        else:
            return body, None



# Licensed to CRATE Technology GmbH ("Crate") under one or more contributor
# license agreements.  See the NOTICE file distributed with this work for
# additional information regarding copyright ownership.  Crate licenses
# this file to you under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.  You may
# obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# License for the specific language governing permissions and limitations
# under the License.
#
# However, if you have executed another commercial license agreement
# with Crate these terms will supersede the license and you may use the
# software solely pursuant to the terms of the relevant commercial agreement.

from argparse import Namespace

from croud.api import Client
from croud.config import get_output_format
from croud.printer import print_response, print_warning
from croud.util import require_confirmation


def consumers_deploy(args: Namespace) -> None:
    """
    The consumer deploy command is deprecated and will be removed in the future.
    There is no explicit deprecation warning being raised, as the API endpoint is
    providing the deprecation message.
    """
    body = {
        "cluster_id": args.cluster_id,
        "config": {
            "connection_string": args.eventhub_dsn,
            "consumer_group": args.eventhub_consumer_group,
            "consumer_lease_container": args.lease_storage_container,
            "lease_storage_connection_string": args.lease_storage_dsn,
        },
        "instances": args.num_instances,
        "name": args.consumer_name,
        "product_name": args.product_name,
        "product_tier": args.tier,
        "project_id": args.project_id,
        "table_name": args.consumer_table,
        "table_schema": args.consumer_schema,
    }
    client = Client.from_args(args)
    data, errors = client.post("/api/v2/consumers/", body=body)
    print_response(
        data=data,
        errors=errors,
        keys=[
            "cluster_id",
            "id",
            "instances",
            "name",
            "product_name",
            "product_tier",
            "project_id",
            "table_name",
            "table_schema",
        ],
        success_message=(
            "Consumer deployed. It may take a few minutes to complete the changes."
        ),
        output_fmt=get_output_format(args),
    )


def consumers_list(args: Namespace) -> None:
    print_warning(
        "The consumers list command is deprecated and will be removed in the future."
    )
    params = {}
    if args.cluster_id:
        params["cluster_id"] = args.cluster_id
    if args.product_name:
        params["product_name"] = args.product_name
    if args.project_id:
        params["project_id"] = args.project_id

    client = Client.from_args(args)
    data, errors = client.get("/api/v2/consumers/", params=params)
    print_response(
        data=data,
        errors=errors,
        keys=[
            "cluster_id",
            "id",
            "instances",
            "name",
            "product_name",
            "product_tier",
            "project_id",
            "table_name",
            "table_schema",
        ],
        output_fmt=get_output_format(args),
    )


def consumers_edit(args: Namespace) -> None:
    print_warning(
        "The consumers edit command is deprecated and will be removed in the future."
    )
    body = {
        "cluster_id": args.cluster_id,
        "table_name": args.consumer_table,
        "table_schema": args.consumer_schema,
    }
    config = {}
    if args.eventhub_dsn:
        config["connection_string"] = args.eventhub_dsn
    if args.eventhub_consumer_group:
        config["consumer_group"] = args.eventhub_consumer_group
    if args.lease_storage_container:
        config["consumer_lease_container"] = args.lease_storage_container
    if args.lease_storage_dsn:
        config["lease_storage_connection_string"] = args.lease_storage_dsn

    if config:
        body["config"] = config
    client = Client.from_args(args)
    data, errors = client.patch(f"/api/v2/consumers/{args.consumer_id}/", body=body)
    print_response(
        data=data, errors=errors, keys=["id"], output_fmt=get_output_format(args)
    )


@require_confirmation(
    "Are you sure you want to delete the consumer?",
    cancel_msg="Consumer deletion cancelled.",
)
def consumers_delete(args: Namespace) -> None:
    print_warning(
        "The consumers delete command is deprecated and will be removed in the future."
    )
    client = Client.from_args(args)
    data, errors = client.delete(f"/api/v2/consumers/{args.consumer_id}/")
    print_response(
        data=data,
        errors=errors,
        success_message="Consumer deleted.",
        output_fmt=get_output_format(args),
    )



# Licensed to CRATE Technology GmbH ("Crate") under one or more contributor
# license agreements.  See the NOTICE file distributed with this work for
# additional information regarding copyright ownership.  Crate licenses
# this file to you under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.  You may
# obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# License for the specific language governing permissions and limitations
# under the License.
#
# However, if you have executed another commercial license agreement
# with Crate these terms will supersede the license and you may use the
# software solely pursuant to the terms of the relevant commercial agreement.

from unittest import mock

import pytest

from croud.api import Client, RequestMethod
from tests.util import assert_rest, call_command, gen_uuid

pytestmark = pytest.mark.usefixtures("config")

eventhub_dsn = "Endpoint=sb://myhub.servicebus.windows.net/;SharedAccessKeyName=...;SharedAccessKey=...;EntityPath=..."  # noqa
storage_dsn = "DefaultEndpointsProtocol=https;AccountName=...;AccountKey=...;EndpointSuffix=core.windows.net"  # noqa


@mock.patch.object(Client, "request", return_value=["data", "errors"])
def test_consumers_deploy(mock_request):
    project_id = gen_uuid()
    cluster_id = gen_uuid()

    call_command(
        "croud",
        "consumers",
        "deploy",
        "--product-name",
        "eventhub-consumer",
        "--tier",
        "S0",
        "--num-instances",
        "2",
        "--project-id",
        project_id,
        "--cluster-id",
        cluster_id,
        "--consumer-name",
        "my-eventhub-consumer",
        "--consumer-table",
        "raw",
        "--consumer-schema",
        "doc",
        "--eventhub-dsn",
        eventhub_dsn,
        "--eventhub-consumer-group",
        "$Default",
        "--lease-storage-dsn",
        storage_dsn,
        "--lease-storage-container",
        "lease_container",
    )

    assert_rest(
        mock_request,
        RequestMethod.POST,
        "/api/v2/consumers/",
        body={
            "cluster_id": cluster_id,
            "config": {
                "connection_string": eventhub_dsn,
                "consumer_group": "$Default",
                "lease_storage_connection_string": storage_dsn,
                "consumer_lease_container": "lease_container",
            },
            "instances": 2,
            "name": "my-eventhub-consumer",
            "product_name": "eventhub-consumer",
            "product_tier": "S0",
            "project_id": project_id,
            "table_name": "raw",
            "table_schema": "doc",
        },
    )


@mock.patch.object(Client, "request", return_value=({}, None))
def test_consumers_list(mock_request, capsys):
    call_command("croud", "consumers", "list")
    assert_rest(mock_request, RequestMethod.GET, "/api/v2/consumers/", params={})
    _, err = capsys.readouterr()
    assert (
        "The consumers list command is deprecated and will be removed in the future."
        in err
    )


@mock.patch.object(Client, "request", return_value=({}, None))
def test_consumers_list_with_params(mock_request, capsys):
    project_id = gen_uuid()
    cluster_id = gen_uuid()
    call_command(
        "croud",
        "consumers",
        "list",
        "--project-id",
        project_id,
        "--cluster-id",
        cluster_id,
        "--product-name",
        "eventhub-consumer",
    )
    assert_rest(
        mock_request,
        RequestMethod.GET,
        "/api/v2/consumers/",
        params={
            "product_name": "eventhub-consumer",
            "project_id": project_id,
            "cluster_id": cluster_id,
        },
    )
    _, err = capsys.readouterr()
    assert (
        "The consumers list command is deprecated and will be removed in the future."
        in err
    )


@mock.patch.object(Client, "request", return_value=({}, None))
def test_consumers_edit(mock_request, capsys):
    consumer_id = gen_uuid()
    cluster_id = gen_uuid()

    call_command(
        "croud",
        "consumers",
        "edit",
        "--consumer-id",
        consumer_id,
        "--eventhub-dsn",
        eventhub_dsn,
        "--eventhub-consumer-group",
        "$Default",
        "--lease-storage-dsn",
        storage_dsn,
        "--lease-storage-container",
        "lease_container",
        "--consumer-schema",
        "doc",
        "--consumer-table",
        "raw",
        "--cluster-id",
        cluster_id,
    )

    assert_rest(
        mock_request,
        RequestMethod.PATCH,
        f"/api/v2/consumers/{consumer_id}/",
        body={
            "cluster_id": cluster_id,
            "config": {
                "connection_string": eventhub_dsn,
                "consumer_group": "$Default",
                "lease_storage_connection_string": storage_dsn,
                "consumer_lease_container": "lease_container",
            },
            "table_name": "raw",
            "table_schema": "doc",
        },
    )
    _, err = capsys.readouterr()
    assert (
        "The consumers edit command is deprecated and will be removed in the future."
        in err
    )


@mock.patch.object(Client, "request", return_value=(None, {}))
def test_consumers_delete(mock_request, capsys):
    consumer_id = gen_uuid()
    with mock.patch("builtins.input", side_effect=["YES"]) as mock_input:
        call_command("croud", "consumers", "delete", "--consumer-id", consumer_id)
    assert_rest(mock_request, RequestMethod.DELETE, f"/api/v2/consumers/{consumer_id}/")
    mock_input.assert_called_once_with(
        "Are you sure you want to delete the consumer? [yN] "
    )

    _, err_output = capsys.readouterr()
    assert "Success" in err_output
    assert "Consumer deleted." in err_output
    assert (
        "The consumers delete command is deprecated and will be removed in the future."
        in err_output
    )


@mock.patch.object(Client, "request", return_value=(None, {}))
def test_consumers_delete_flag(mock_request, capsys):
    consumer_id = gen_uuid()
    with mock.patch("builtins.input", side_effect=["y"]) as mock_input:
        call_command("croud", "consumers", "delete", "--consumer-id", consumer_id, "-y")
    assert_rest(mock_request, RequestMethod.DELETE, f"/api/v2/consumers/{consumer_id}/")
    mock_input.assert_not_called()

    _, err_output = capsys.readouterr()
    assert "Success" in err_output
    assert "Consumer deleted." in err_output
    assert (
        "The consumers delete command is deprecated and will be removed in the future."
        in err_output
    )


@mock.patch.object(Client, "request", return_value=(None, {}))
def test_consumers_delete_aborted(mock_request, capsys):
    consumer_id = gen_uuid()
    with mock.patch("builtins.input", side_effect=["Nooooo"]) as mock_input:
        call_command("croud", "consumers", "delete", "--consumer-id", consumer_id)
    mock_request.assert_not_called()
    mock_input.assert_called_once_with(
        "Are you sure you want to delete the consumer? [yN] "
    )

    _, err_output = capsys.readouterr()
    assert "Consumer deletion cancelled." in err_output



# Licensed to CRATE Technology GmbH ("Crate") under one or more contributor
# license agreements.  See the NOTICE file distributed with this work for
# additional information regarding copyright ownership.  Crate licenses
# this file to you under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.  You may
# obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# License for the specific language governing permissions and limitations
# under the License.
#
# However, if you have executed another commercial license agreement
# with Crate these terms will supersede the license and you may use the
# software solely pursuant to the terms of the relevant commercial agreement.

from unittest import mock

import pytest

from croud.api import Client, RequestMethod
from tests.util import assert_rest, call_command, gen_uuid

pytestmark = pytest.mark.usefixtures("config")


@mock.patch.object(Client, "request", return_value=({}, None))
def test_clusers_list(mock_request):
    call_command("croud", "clusters", "list")
    assert_rest(mock_request, RequestMethod.GET, "/api/v2/clusters/", params={})


@mock.patch.object(Client, "request", return_value=({}, None))
def test_clusers_list_with_project_id(mock_request):
    project_id = gen_uuid()
    call_command("croud", "clusters", "list", "--project-id", project_id)
    assert_rest(
        mock_request,
        RequestMethod.GET,
        "/api/v2/clusters/",
        params={"project_id": project_id},
    )


@mock.patch.object(Client, "request", return_value=({}, None))
def test_clusers_deploy(mock_request):
    project_id = gen_uuid()
    call_command(
        "croud",
        "clusters",
        "deploy",
        "--product-name",
        "cratedb.az1",
        "--tier",
        "xs",
        "--unit",
        "1",
        "--project-id",
        project_id,
        "--cluster-name",
        "crate_cluster",
        "--version",
        "3.2.5",
        "--username",
        "foobar",
        "--password",
        "s3cr3t!",
    )
    assert_rest(
        mock_request,
        RequestMethod.POST,
        "/api/v2/clusters/",
        body={
            "crate_version": "3.2.5",
            "name": "crate_cluster",
            "password": "s3cr3t!",
            "product_name": "cratedb.az1",
            "product_tier": "xs",
            "product_unit": 1,
            "project_id": project_id,
            "username": "foobar",
            "channel": "stable",
        },
    )


@mock.patch.object(Client, "request", return_value=({}, None))
def test_clusers_deploy_no_unit(mock_request):
    project_id = gen_uuid()
    call_command(
        "croud",
        "clusters",
        "deploy",
        "--product-name",
        "cratedb.az1",
        "--tier",
        "xs",
        "--project-id",
        project_id,
        "--cluster-name",
        "crate_cluster",
        "--version",
        "3.2.5",
        "--username",
        "foobar",
        "--password",
        "s3cr3t!",
    )
    assert_rest(
        mock_request,
        RequestMethod.POST,
        "/api/v2/clusters/",
        body={
            "crate_version": "3.2.5",
            "name": "crate_cluster",
            "password": "s3cr3t!",
            "product_name": "cratedb.az1",
            "product_tier": "xs",
            "project_id": project_id,
            "username": "foobar",
            "channel": "stable",
        },
    )


@mock.patch.object(Client, "request", return_value=({}, None))
def test_clusers_deploy_nightly(mock_request):
    project_id = gen_uuid()
    call_command(
        "croud",
        "clusters",
        "deploy",
        "--product-name",
        "cratedb.az1",
        "--tier",
        "xs",
        "--unit",
        "1",
        "--project-id",
        project_id,
        "--cluster-name",
        "crate_cluster",
        "--version",
        "nightly-4.1.0-20190712",
        "--username",
        "foobar",
        "--password",
        "s3cr3t!",
        "--channel",
        "nightly",
    )
    assert_rest(
        mock_request,
        RequestMethod.POST,
        "/api/v2/clusters/",
        body={
            "crate_version": "nightly-4.1.0-20190712",
            "name": "crate_cluster",
            "password": "s3cr3t!",
            "product_name": "cratedb.az1",
            "product_tier": "xs",
            "project_id": project_id,
            "username": "foobar",
            "channel": "nightly",
            "product_unit": 1,
        },
    )


@mock.patch.object(Client, "request", return_value=({}, None))
def test_clusers_scale(mock_request):
    unit = 1
    cluster_id = gen_uuid()
    call_command(
        "croud", "clusters", "scale", "--cluster-id", cluster_id, "--unit", "1"
    )
    assert_rest(
        mock_request,
        RequestMethod.PUT,
        f"/api/v2/clusters/{cluster_id}/scale/",
        body={"product_unit": unit},
    )


@mock.patch.object(Client, "request", return_value=({}, None))
def test_clusers_upgrade(mock_request):
    version = "3.2.6"
    cluster_id = gen_uuid()
    call_command(
        "croud", "clusters", "upgrade", "--cluster-id", cluster_id, "--version", version
    )
    assert_rest(
        mock_request,
        RequestMethod.PUT,
        f"/api/v2/clusters/{cluster_id}/upgrade/",
        body={"crate_version": version},
    )


@mock.patch.object(Client, "request", return_value=(None, {}))
def test_clusters_delete(mock_request, capsys):
    cluster_id = gen_uuid()
    with mock.patch("builtins.input", side_effect=["yes"]) as mock_input:
        call_command("croud", "clusters", "delete", "--cluster-id", cluster_id)
    assert_rest(mock_request, RequestMethod.DELETE, f"/api/v2/clusters/{cluster_id}/")
    mock_input.assert_called_once_with(
        "Are you sure you want to delete the cluster? [yN] "
    )

    _, err_output = capsys.readouterr()
    assert "Success" in err_output
    assert "Cluster deleted." in err_output


@mock.patch.object(Client, "request", return_value=(None, {}))
def test_clusters_delete_flag(mock_request, capsys):
    cluster_id = gen_uuid()
    with mock.patch("builtins.input", side_effect=["y"]) as mock_input:
        call_command("croud", "clusters", "delete", "--cluster-id", cluster_id, "-y")
    assert_rest(mock_request, RequestMethod.DELETE, f"/api/v2/clusters/{cluster_id}/")
    mock_input.assert_not_called()

    _, err_output = capsys.readouterr()
    assert "Success" in err_output
    assert "Cluster deleted." in err_output


@mock.patch.object(Client, "request", return_value=(None, {}))
def test_clusters_delete_aborted(mock_request, capsys):
    cluster_id = gen_uuid()
    with mock.patch("builtins.input", side_effect=["Nooooo"]) as mock_input:
        call_command("croud", "clusters", "delete", "--cluster-id", cluster_id)
    mock_request.assert_not_called()
    mock_input.assert_called_once_with(
        "Are you sure you want to delete the cluster? [yN] "
    )

    _, err_output = capsys.readouterr()
    assert "Cluster deletion cancelled." in err_output
